package se.chalmers.paragon;

import java.util.HashSet;

/**
 * <p>
 * Represents a lockstate for the associated lock.
 * </p>
 * 
 * @author Paragon Developers (http://www.cse.chalmers.se/paragon)
 * 
 */
public class LockState {

    HashSet<ActorList> currentState = new HashSet<ActorList>();
    private Lock lock;

    protected LockState(Lock lock) {
        this.lock = lock;
    }

    /**
     * Adds the actors to the state of open locks (if they are not in this state
     * already). Throws {@link se.chalmers.paragon.LockArityError} in case the
     * argument is of the wrong length.
     * 
     * @param actors
     * 
     * @return true if the lock was not already opened
     */
    public boolean open(ConcreteActor... actors) {

        if (actors.length != this.lock.getArity())
            throw new LockArityError(actors, this.lock);

        return currentState.add(new ActorList(actors));

    }

    /**
     * Removes this pair of actors from the state of open locks. Throws
     * {@link se.chalmers.paragon.LockArityError} in case the argument is of the
     * wrong length.
     * 
     * @param actors
     */
    public boolean close(ConcreteActor... actors) {

        if (actors.length != this.lock.getArity())
            throw new LockArityError(actors, this.lock);

        return currentState.remove(new ActorList(actors));

    }

    /**
     * Checks if this pair of actors is in the state of open lock. Throws
     * {@link se.chalmers.paragon.LockArityError} in case the argument is of the
     * wrong length.
     * 
     * @param actors
     * @return True if the provided list of actors is in the lock state, false
     *         otherwise
     */
    public boolean isOpen(ConcreteActor... actors) {

        if (actors.length != this.lock.getArity())
            throw new LockArityError(actors, this.lock);

        return currentState.contains(new ActorList(actors));
    }

}
